﻿using IMAPI2FS;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using ADODB;
using System.Collections;
using System.IO;

namespace MyISO.Model.Global
{
    /// <summary>
    /// Cette classe fournit les fonctionnalités nécessaires à la création d'une image ISO (bootable ou non).
    /// This class provides the functionalities which are useful to create an ISO image (bootable or not).
    /// </summary>
    /// <author>
    /// <name>
    /// Grégory MARTIN
    /// </name>
    /// <website>
    /// http://www.gmdevelopment.info
    /// </website>
    /// </author>
    /// <Acknowledgement>
    /// This code is largely inspired from this article:http://www.codeproject.com/Articles/23653/How-to-Create-Optical-File-Images-using-IMAPIv
    /// The bootable image support is inspired from this MSDN page:
    /// https://msdn.microsoft.com/en-us/library/windows/desktop/aa364813%28v=vs.85%29.aspx
    /// </Acknowledgement>
    class ISO9660
    {
        #region Flags + DllImport
        [Flags]
        public enum StgmConstants
        {
            STGM_READ = 0x0,
            STGM_WRITE = 0x1,
            STGM_READWRITE = 0x2,
            STGM_SHARE_DENY_NONE = 0x40,
            STGM_SHARE_DENY_READ = 0x30,
            STGM_SHARE_DENY_WRITE = 0x20,
            STGM_SHARE_EXCLUSIVE = 0x10,
            STGM_PRIORITY = 0x40000,
            STGM_CREATE = 0x1000,
            STGM_CONVERT = 0x20000,
            STGM_FAILIFTHERE = 0x0,
            STGM_DIRECT = 0x0,
            STGM_TRANSACTED = 0x10000,
            STGM_NOSCRATCH = 0x100000,
            STGM_NOSNAPSHOT = 0x200000,
            STGM_SIMPLE = 0x8000000,
            STGM_DIRECT_SWMR = 0x400000,
            STGM_DELETEONRELEASE = 0x4000000
        }

        [DllImport("shlwapi.dll", CharSet = CharSet.Unicode, ExactSpelling = true, PreserveSig = false, EntryPoint = "SHCreateStreamOnFileW")]
        private static extern void SHCreateStreamOnFile(string fileName, uint mode, ref System.Runtime.InteropServices.ComTypes.IStream stream);
        #endregion

        #region Fields
        private string _Path = String.Empty;
        private string _ISOPath = String.Empty;
        private string _VolumeName = String.Empty;
        private bool _Bootable = true;
        private string _BootFile = string.Empty;
        #endregion

        #region Delegates
        private delegate void DCreateISO();
        private DCreateISO _CreateFile;
        private DCreateISO CreateFile
        {
            get
            {
                return _CreateFile = _CreateFile ?? new DCreateISO(CreateISO);
            }
        }
        #endregion

        #region Properties
        public bool InProgress
        {
            get;
            set;
        }

        public string Path
        {
            get { return _Path; }
            set { _Path = value; }
        }

        public string ISOPath
        {
            get { return _ISOPath; }
            set { _ISOPath = value; }
        }

        public string VolumeName
        {
            get { return _VolumeName; }
            set { _VolumeName = value; }
        }
        #endregion

        #region Constructors
        public ISO9660() { }
        #endregion
        /// <summary>
        /// Cette méthode crée une image ISO qui a pour emplacement this.ISOPath,
        /// pour nom de volume this.VolumeName, pour BootFile this.BootFile (peut être string.Empty
        /// indépendamment de la valeur de this.IsBootable (true|false)). Si this.IsBootable == true
        /// alors this.BootFile est ignoré. À la fin de la méthode, on a un fichier ISO qui contient
        /// l'arborescence de fichiers à l'eplacement this.Path. Si this.IsBootable == true, le fichier
        /// ISO est bootable, sinon il ne l'est pas.
        /// 
        /// This method creates an ISO image which has the following characteristics:
        /// - the file is stored at this.ISOPath;
        /// - the volume name of the ISO image is this.VolumeName;
        /// - the ISO image is bootable if and only if this.IsBootable == true;
        /// - this.BootFile must have a value not null and not string.Empty if this.BootFile == true;
        /// If this.BootFile has a value and this.IsBootable is false then this.BootFile is ignored.
        /// </summary>
        public void CreateISO()
        {
            //InProgress = true;
            MsftFileSystemImage iso = new MsftFileSystemImage();
            iso.ChooseImageDefaultsForMediaType(IMAPI_MEDIA_PHYSICAL_TYPE.IMAPI_MEDIA_TYPE_DISK);
            iso.VolumeName = this._VolumeName;
            ADODB.Stream BootStream = new ADODB.Stream();
            if (_Bootable)
            {
                //https://msdn.microsoft.com/en-us/library/windows/desktop/aa364813%28v=vs.85%29.aspx
                IMAPI2FS.BootOptions BootOptions = new IMAPI2FS.BootOptions();
                BootOptions.PlatformId = PlatformId.PlatformX86;
                BootOptions.Manufacturer = "Microsoft";
                BootOptions.Emulation = EmulationType.EmulationNone;
                BootStream.Open();
                BootStream.Type = StreamTypeEnum.adTypeBinary;
                BootStream.LoadFromFile(_BootFile);
                BootOptions.AssignBootImage((FsiStream)BootStream);
                iso.BootImageOptions = BootOptions;
            }
            iso.FileSystemsToCreate = FsiFileSystems.FsiFileSystemUDF;
            iso.Root.AddTree(this._Path, false);
            IFileSystemImageResult fsimage = iso.CreateResultImage();
            FsiStream isoStream = fsimage.ImageStream;
            System.Runtime.InteropServices.ComTypes.IStream stream = null;
            if (isoStream != null)
            {
                tagSTATSTG stat;
                isoStream.Stat(out stat, 0x1);
                SHCreateStreamOnFile(this._ISOPath, (uint)(StgmConstants.STGM_WRITE | StgmConstants.STGM_CREATE | StgmConstants.STGM_SHARE_DENY_WRITE), ref stream);
                if (stream != null)
                {
                    IMAPI2FS._ULARGE_INTEGER inBytes, outBytes;
                    try
                    {
                        isoStream.RemoteCopyTo((FsiStream)stream, stat.cbSize, out inBytes, out outBytes);
                    }
                    catch (Exception ex)
                    {
                        LastException = ex;
                    }
                    finally
                    {
                        stream.Commit(0);
                        Marshal.FinalReleaseComObject(stream);
                        MessageBox.Show("ISO file created. Releasing lock on files.");
                        //ReleaseLockOnFiles(iso.Root);
                        MessageBox.Show("Files released");
                        Marshal.FinalReleaseComObject(isoStream);
                        Marshal.FinalReleaseComObject(fsimage);
                        Marshal.FinalReleaseComObject(iso);
                        Marshal.FinalReleaseComObject(BootStream);
                    }
                }
                else
                {
                    Marshal.FinalReleaseComObject(isoStream);
                    Marshal.FinalReleaseComObject(fsimage);
                    Marshal.FinalReleaseComObject(iso);
                }
            }
            else
            {
                Marshal.FinalReleaseComObject(fsimage);
                Marshal.FinalReleaseComObject(iso);
            }
            InProgress = false;
        }

        /// <summary>
        /// Cette méthode libère les fichiers sur lesquels la méthode CreateISO() a mis un verrou.
        /// Cette méthode est appelée récursivement sur toute l'arborescence de fichiers et répertoires
        /// à l'emplacement this.Path.
        /// 
        /// This method releases the files which were locked by the CreateISO() method.
        /// This method is recursively called on each subdirectory in this.Path.
        /// </summary>
        /// <param name="RootItem"></param>
        private void ReleaseLockOnFiles(IFsiDirectoryItem RootItem)
        {
            if (RootItem != null)
            {
                IEnumerator items = RootItem.GetEnumerator();
                DirectoryInfo dirInfo = null;
                foreach (IFsiItem item in RootItem)
                {
                    dirInfo = new DirectoryInfo(this._Path + item.FullPath);
                    if (!dirInfo.Exists)
                    {
                        IFsiFileItem file = (IFsiFileItem)item;
                        try
                        {
                            //IFsiFileItem file = (IFsiFileItem)item;
                            FsiStream stream = file.Data;
                            IntPtr IUnknownForObject = Marshal.GetIUnknownForObject(stream);
                            var i = Marshal.AddRef(IUnknownForObject);
                            while (i > 0)
                            {
                                i = Marshal.Release(IUnknownForObject);
                            }
                            Marshal.FinalReleaseComObject(stream);
                        }
                        //catch (InvalidCastException CastEx)
                        //{
                        //    try
                        //    {
                        //        IFsiDirectoryItem dir = (IFsiDirectoryItem)items.Current;
                        //        ReleaseLockOnFiles(dir);
                        //    }
                        //    catch (InvalidCastException castex)
                        //    {
                        //        MessageBox.Show("nor file nor dir");
                        //    }
                        //}
                        catch (Exception ex)
                        {
                            MessageBox.Show(ex.Message);
                        }
                    }
                    else
                        ReleaseLockOnFiles((IFsiDirectoryItem) item);
                }
            }
        }

        public Exception LastException { get; set; }

        /// <summary>
        /// Cette fonction lance un thread non bloquant qui exécute le délégué CreateFile.
        /// </summary>
        /// <param name="Callback">La fonction à exécuter à la fin de cette méthode. The function to be executed at the end of this method</param>
        /// <param name="IsBootable"></param>
        /// <param name="BootFile">
        /// Un string qui contient l'emplacement du fichier de boot à inclure dans le fichier ISO. 
        /// A string that contains the place where the boot file to include in the ISO file is stored.
        /// </param>
        /// <seealso cref="MyISO.Model.Global.ISO9660.CreateFile"/>
        internal void StartCreation(AsyncCallback Callback, bool IsBootable, string BootFile)
        {
            this._Bootable = IsBootable;
            this._BootFile = BootFile;
            this.CreateFile.BeginInvoke(Callback, null);
        }
    }
}
