using GalaSoft.MvvmLight;
using GalaSoft.MvvmLight.Command;
using System;
using System.Windows.Data;
using System.Windows.Forms;
using System.Windows.Input;
using MyISO.Model.Client;
using System.Globalization;

namespace MyISO.ViewModel
{
    /// <summary>
    /// Cette classe contient des proprits sur lesquelles la vue principale peut faire un databinding.
    /// This class contains properties that the main View can data bind to.
    /// <para>
    /// Use the <strong>mvvminpc</strong> snippet to add bindable properties to this ViewModel.
    /// </para>
    /// <para>
    /// You can also use Blend to data bind with the tool's support.
    /// </para>
    /// <para>
    /// See http://www.galasoft.ch/mvvm
    /// </para>
    /// </summary>
    /// <author>
    /// <name>
    /// Grgory MARTIN
    /// </name>
    /// <website>
    /// http://www.gmdevelopment.info
    /// </website>
    /// </author>
    public class MainViewModel : ViewModelBase
    {
        #region Properties
        private ISO9660Client _ISO = new ISO9660Client();

        internal ISO9660Client ISO
        {
            get { return _ISO; }
            set { _ISO = value; }
        }

        public volatile bool _InProgress = false;
        public bool InProgress
        {
            get
            {
                return this._ISO.InProgress;
                //return this._InProgress;
            }
            set
            {
                //this._InProgress = value;
                this._ISO.InProgress = value;
                RaisePropertyChanged(() => InProgress);
            }
        }

        private bool _IsBootable;

        public bool IsBootable
        {
            get
            {
                return _IsBootable;
            }

            set
            {
                _IsBootable = value;
            }
        }

        private string _BootFile = @"D:\Backup\Windows\boot\etfsboot.com";

        public string BootFile
        {
            get
            {
                return _BootFile;
            }

            set
            {
                _BootFile = value;
            }
        }

        public string Path
        {
            get
            {
                return this._ISO.Path;
            }

            set
            {
                this._ISO.Path = value;
                RaisePropertyChanged(() => Path);
            }
        }

        public string ISOPath
        {
            get
            {
                return this._ISO.ISOPath;
            }

            set
            {
                this._ISO.ISOPath = value;
                RaisePropertyChanged(() => ISOPath);
            }
        }

        public string VolumeName
        {
            get
            {
                return this._ISO.VolumeName;
            }

            set
            {
                this._ISO.VolumeName = value;
                RaisePropertyChanged(() => VolumeName);
            }
        }
        #endregion

        #region Commands
        #region Creation
        /// <summary>
        /// Cette commande excute StartCreation si CanCreateISO renvoie true.
        /// This command executes StartCreation if CanCreateISO returns true.
        /// <see cref="CanCreateISO"/>
        /// <seealso cref="StartCreation"/>
        /// </summary>
        private RelayCommand _RcCreateISO;
        public RelayCommand RcCreateISO
        {
            get { return _RcCreateISO = _RcCreateISO ?? new RelayCommand(StartCreation, CanCreateISO); }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <returns>!this.InProgress & this.VolumeName != string.Empty && this.ISOPath != string.Empty && this.Path != string.Empty</returns>
        private bool CanCreateISO()
        {
            bool can;
            lock (this._Lock)
            {
                can = !this.InProgress;
            }
            can &= this.VolumeName != string.Empty && this.ISOPath != string.Empty && this.Path != string.Empty;
            return can;
        }


        private AsyncCallback _Callback;

        /// <summary>
        /// La mthode  excuter aprs la cration d'une image ISO.
        /// <see cref="SetInProgressFalse(IAsyncResult)"/>
        /// </summary>
        public AsyncCallback Callback
        {
            get { return _Callback = _Callback ?? new AsyncCallback(SetInProgressFalse); }
            set { _Callback = value; }
        }

        private object _Lock = new object();
        private void SetInProgressFalse(IAsyncResult ar)
        {
            lock (this._Lock)
            {
                this.InProgress = false;
            }
            this.RcBrowse.RaiseCanExecuteChanged();
            this.RcCreateISO.RaiseCanExecuteChanged();
        }

        /// <summary>
        /// Met this.InProgress  true et appelle this._ISO.StartCreation(Callback, IsBootable, BootFile).
        /// Cet appel lance un thread pour ne pas bloquer l'interface graphique.
        /// Le thread cre un fichier ISO  l'emplacement this.ISOPath et ce fichier 
        /// contient l'arborescence de fichiers  l'emplacement this.Path. De plus, l'ISO est bootable si IsBootable == true.
        /// Pas de vrification sur le fichier bootable.
        /// 
        /// Sets this.InProgress = true and calls this._ISO.StartCreation(Callback, IsBootable, BootFile).
        /// This call launches a thread to avoid a not responding GUI.
        /// This thread creates an ISO file at this.ISOPath and this file contains the file tree at
        /// this.Path. Moreover, this ISO file is bootable if this.IsBootable == true.
        /// <see cref="InProgress"/>
        /// <seealso cref="MyISO.Model.Client.ISO9660Client.StartCreation(AsyncCallback, bool, string)"/>
        /// </summary>
        private void StartCreation()
        {
            this.InProgress = true;
            this._ISO.StartCreation(Callback, IsBootable, BootFile);
        }
        #endregion

        #region Browse
        private RelayCommand _RcBrowse;
        /// <summary>
        /// Ouvre une bote de dialogue de slection du dossier  intgrer  l'image ISO.
        /// Opens a folder browser dialog to choose a directory to include in the ISO file.
        /// </summary>
        public RelayCommand RcBrowse
        {
            get { return _RcBrowse = _RcBrowse ?? new RelayCommand(Browse, CanBrowse); }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <returns>!this.InProgress</returns>
        private bool CanBrowse()
        {
            lock (this._Lock)
            {
                return !this.InProgress;
            }
        }

        private void Browse()
        {
            FolderBrowserDialog folderDial = new FolderBrowserDialog();
            DialogResult res = folderDial.ShowDialog();
            if (res == DialogResult.OK)
            {
                this.Path = folderDial.SelectedPath;
            }
        }
        #endregion
        #endregion
        /// <summary>
        /// Initializes a new instance of the MainViewModel class.
        /// </summary>
        public MainViewModel()
        {
            ////if (IsInDesignMode)
            ////{
            ////    // Code runs in Blend --> create design time data.
            ////}
            ////else
            ////{
            ////    // Code runs "for real"
            ////}
        }
    }
}